import numpy as np
from .core import to_44, compose_rt, normalize

def sincos(a, degrees=True):
    """
    sin and cos of an angle
    :param a: angle
    :param degrees: if True, a is in degrees
    """
    if degrees:
        a = np.deg2rad(a)
    return np.sin(a), np.cos(a)

def translate(t):
    """
    creates a translation matrix from a translation vector
    :param t: translation vector or batch of translation vectors
    :return: 4x4 translation matrix
    """
    if t.shape[-1] != 3:
        raise ValueError("translation vector must be 3d")
    if t.ndim == 1:
        t = t[None, :]
    mat = np.concatenate((np.eye(3)[None, :, :], t[:, :, None]), axis=-1)
    return to_44(mat)

def scale(s):
    """
    creates a scaling matrix from a scaling vector
    :param s: scaling vector or batch of scaling vectors
    :return: nx4x4 scaling matrix
    """
    if s.shape[-1] != 3:
        raise ValueError("translation vector must be 3d")
    if s.ndim == 1:
        s = s[None, :]
    mat = np.diag(s)
    return to_44(mat)

def rotate(a, r):
    """
    creates a rotation matrix from an angle a and an axis of rotation r
    """
    s, c = sincos(a)
    r = normalize(r)
    nc = 1 - c
    x, y, z = r
    return np.array([[x*x*nc +   c, x*y*nc - z*s, x*z*nc + y*s, 0],
                      [y*x*nc + z*s, y*y*nc +   c, y*z*nc - x*s, 0],
                      [x*z*nc - y*s, y*z*nc + x*s, z*z*nc +   c, 0],
                      [           0,            0,            0, 1]])
def rotx(a, degrees=True):
    """
    creates a homogeneous 3D rotation matrix around the x axis
    a: angle
    degrees: if True, a is in degrees, else radians
    """
    s, c = sincos(a, degrees)
    return np.array([[1,0,0,0],
                      [0,c,-s,0],
                      [0,s,c,0],
                      [0,0,0,1]])

def roty(a, degrees=True):
    """
    creates a homogeneous 3D rotation matrix around the y axis
    a: angle
    degrees: if True, a is in degrees, else radians
    """
    s, c = sincos(a, degrees)
    return np.array([[c,0,s,0],
                      [0,1,0,0],
                      [-s,0,c,0],
                      [0,0,0,1]])

def rotz(a, degrees=True):
    """
    creates a homogeneous 3D rotation matrix around the z axis
    a: angle
    degrees: if True, a is in degrees, else radians
    """
    s, c = sincos(a, degrees)
    return np.array([[c,-s,0,0],
                      [s,c,0,0],
                      [0,0,1,0],
                      [0,0,0,1]])

def find_rigid_transform(A, B):
    """
    finds best 3d rigid transformation between pc a and pc b (in the least squares sense) based on "Least-Squares Rigid Motion Using SVD"
    :param A: point cloud a
    :param B: point cloud b
    :return: W such that W@A = B
    """
    centroid_A = np.mean(A, axis=0)
    centroid_B = np.mean(B, axis=0)

    A_mean = A - centroid_A
    B_mean = B - centroid_B

    H = A_mean.T @ B_mean
    U, S, Vt = np.linalg.svd(H)

    flip = np.linalg.det(Vt.T @ U.T)
    ones = np.identity(len(Vt))
    ones[-1, -1] = flip
    R = Vt.T @ ones @ U.T
    t = centroid_B - R @ centroid_A
    return compose_rt(R[None, :], t[None, :], square=True)

def find_affine_transformation(A, B):
    """
    finds best 3d affine transformation between pc a and pc b (in the least squares sense)
    :param A: point cloud a
    :param B: point cloud b
    :return: W such that W@A = B
    """
    # assemble A matrix for least squares
    assert len(A) >= 4
    new_A = np.c_[A, np.ones(len(A))]
    new_A = np.repeat(np.tile(new_A, (1, 3)), 3, axis=0)
    # now A is 3nx12
    # zero columns depending on index of point
    new_A[::3, 4:] = 0
    new_A[1::3, :4] = 0
    new_A[1::3, 8:] = 0
    new_A[2::3, :8] = 0
    new_B = B.flatten()
    if len(A) == 4:
        #solve linear system
        W = np.linalg.inv(new_A) @ new_B
    else:
        W = np.linalg.lstsq(new_A, new_B, rcond=None)[0]
    return to_44(W.reshape(3, 4))

def decompose_affine(A44):
    """
    Decompose 4x4 homogenous affine matrix into parts.
    The parts are translations, rotations, zooms, shears.
    Returns
    -------
    T : array, shape (3,)
       Translation vector
    R : array shape (3,3)
        rotation matrix
    Z : array, shape (3,)
       scale vector.  May have one negative zoom to prevent need for negative
       determinant R matrix above
    S : array, shape (3,)
       Shear vector, such that shears fill upper triangle above
       diagonal to form shear matrix (type ``striu``).
    Implementation of "Decomposing a matrix into simple transformations".
    """
    A44 = np.asarray(A44)
    T = A44[:-1,-1]
    RZS = A44[:-1,:-1]
    # compute scales and shears
    M0, M1, M2 = np.array(RZS).T
    # extract x scale and normalize
    sx = np.sqrt(np.sum(M0**2))
    M0 /= sx
    # orthogonalize M1 with respect to M0
    sx_sxy = np.dot(M0, M1)
    M1 -= sx_sxy * M0
    # extract y scale and normalize
    sy = np.sqrt(np.sum(M1**2))
    M1 /= sy
    sxy = sx_sxy / sx
    # orthogonalize M2 with respect to M0 and M1
    sx_sxz = np.dot(M0, M2)
    sy_syz = np.dot(M1, M2)
    M2 -= (sx_sxz * M0 + sy_syz * M1)
    # extract z scale and normalize
    sz = np.sqrt(np.sum(M2**2))
    M2 /= sz
    sxz = sx_sxz / sx
    syz = sy_syz / sy
    # Reconstruct rotation matrix, ensure positive determinant
    Rmat = np.array([M0, M1, M2]).T
    if np.linalg.det(Rmat) < 0:
        sx *= -1
        Rmat[:,0] *= -1
    return T, Rmat, np.array([sx, sy, sz]), np.array([sxy, sxz, syz])


import numpy as np
import torch
import cv2
from .gsoup_io import save_image, save_images, load_images, load_image
from .core import to_8b
from .image import interpolate_multi_channel, change_brightness
from pathlib import Path
from scipy.interpolate import LinearNDInterpolator

def warp_image(backward_map, desired_image, cam_wh=None, output_path=None):
    """
    given a 2D dense map of corresponding pixels between projector and camera, computes a warped image such that if projected, the desired image appears when observed using camera
    :param backward_map: 2D dense mapping between pixels from projector 2 camera (proj_h x proj_w x 2) uint32
    :param desired_image: path to desired image from camera, or float np array channels last (cam_h x cam_w x 3) uint8
    :param cam_wh: device1 (width, height) as tuple, if not supplied assumes desired_image is in the correct dimensions in relation to backward_map
    :param output_path: path to save warped image to
    :return: warped image (proj_h x proj_w x 3) uint8
    """
    if backward_map.ndim != 3:
        raise ValueError("backward_map must be 3D")
    if backward_map.shape[-1] != 2:
        raise ValueError("backward_map must have shape (cam_h, cam_w, 2)")
    if type(desired_image) == np.ndarray:
        if desired_image.ndim != 3:
            raise ValueError("desired_image must be 3D")
    else:
        if cam_wh is not None:
            desired_image = load_image(desired_image, resize_wh=cam_wh)
        else:
            desired_image = load_image(desired_image)
    warpped = desired_image[(backward_map[..., 0], backward_map[..., 1])]
    if output_path is not None:
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        save_image(warpped, output_path)
    return warpped

def compute_backward_map(proj_wh, forward_map, foreground, output_dir=None, debug=False): 
    """
    computes the inverse map of forward_map by piece-wise interpolating a triangulated version of it.
    :param proj_wh: projector (width, height) as a tuple
    :param forward_map: forward map as a numpy array of shape (height, width, 2) of type int32
    :param output_dir: directory to save the inverse map to
    :param debug: if True, saves a visualization of the inverse map to output_dir
    :return: inverse map as a numpy array of shape (projector_height, projector_width, 2) of type int32
    """   
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
    data = np.argwhere(foreground)
    points = forward_map[foreground]
    interp = LinearNDInterpolator(points, data, fill_value=0.0)
    X, Y = np.meshgrid(np.arange(proj_wh[1]), np.arange(proj_wh[0]))
    result = interp(X, Y).transpose(1, 0, 2)  # eww
    if output_dir:
        np.save(Path(output_dir, "backward_map.npy"), result)
        if debug:
            result_normalized = result / np.array([forward_map.shape[0], forward_map.shape[1]])
            result_normalized_8b = to_8b(result_normalized)
            result_normalized_8b_3c = np.concatenate((result_normalized_8b, np.zeros_like(result_normalized_8b[..., :1])), axis=-1)
            save_image(result_normalized_8b_3c, Path(output_dir, "backward_map.png"))
    return result.round().astype(np.uint32)

def naive_color_compensate(target_image, all_white_image, all_black_image, cam_width, cam_height, brightness_decrease=-127, projector_gamma=2.2, output_path=None, debug=False):
    """
    color compensate a projected image such that it appears closer to a target image from the perspective of a camera
    loosly based on "Embedded entertainment with smart projectors"
    :param target_image the desired image path from the perspective of the camera
    :param all_white_image a path to picture taken by camera when projector had all pixels fully on (float32)
    :param all_black_image a path to picture taken by camera when projector had all pixels fully off (float32)
    :param cam_width camera image width
    :param cam_height camera image height
    :param brightness_decrease a hyper parameter controlling how much the total brightness is decreased. without this, the result is saturated because of dividing by small numbers
    :param projector_gamma under normal circumstances the projector does not actually output linear values, so we need to compensate for that
    :output_path if passed, result will be saved to this path
    :debug if true, will save debug info into output_path parent directory
    """
    if output_path is not None:
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)
    target_image = load_image(target_image, to_float=True, resize_wh=(cam_width, cam_height))[..., :3]
    target_image = change_brightness(target_image, brightness_decrease)
    if debug:
        save_image(target_image, Path(output_path.parent, "decrease_brightness.png"))
    all_white_image = load_image(all_white_image, to_float=True, resize_wh=(cam_width, cam_height))
    all_black_image = load_image(all_black_image, to_float=True, resize_wh=(cam_width, cam_height))
    compensated = (target_image - all_black_image) / all_white_image
    compensated = np.power(compensated, (1/projector_gamma))
    compensated = np.nan_to_num(compensated, nan=0.0, posinf=0.0, neginf=0.0)
    compensated = np.clip(compensated, 0, 1)
    if output_path:
        save_image(compensated, output_path)
    return compensated

def calibrate_procam(proj_height, proj_width, graycode_step, capture_dir, 
                     chess_vert=10, chess_hori=7,
                     bg_threshold=10, chess_block_size=10.0, verbose=True,
                     output_dir=None, debug=False):
    """
    calibrates a projection-camera pair using local homographies
    based on "Simple, accurate, and robust projector-camera calibration."
    :param proj_height projector pixel height
    :param proj_width projector pixel width
    :param chess_vert number of cross points of chessboard in vertical direction (not including the border, i.e. internal corners)
    :param chess_hori number of cross points of chessboard in horizontal direction (not including the border, i.e. internal corners)
    :param graycode_step factor used to downsample the graycode images (see generate_gray_code)
    :param capture_dir directory containing the captured images when gray code patterns were projected, assumes structure as follows:
        capture_dir
            - folder1
                - 0000.png
                - 0001.png
                - ...
            - folder2
                - 0000.png
                - ...
    :param bg_threshold threshold for detecting foreground
    :param chess_block_size size of blocks of chessboard in real world in any unit of measurement (will only effect the translation componenet between camera and projector)
    :param verbose if true, will print out the calibration results
    :param output_dir will save debug results to this directory
    :param debug if true, will save debug info into output_dir
    :return camera intrinsics, camera extrinsics, projector intrinsics, projector extrinsics, cam_proj_rmat, cam_proj_tvec
    """
    proj_shape = (proj_height, proj_width)
    chess_shape = (chess_vert, chess_hori)
    gc_step = graycode_step
    capture_dir = Path(capture_dir)
    if not capture_dir.exists():
        raise FileNotFoundError("capture_dir was not found")
    dirnames = sorted(capture_dir.glob('*'))
    if len(dirnames) == 0:
        raise FileNotFoundError("capture_dir contains no subfolders")
    used_dirnames = []
    gc_fname_lists = []
    for dname in dirnames:
        gc_fnames = sorted(dname.glob('*'))
        if len(gc_fnames) == 0:
            continue
        used_dirnames.append(str(dname))
        gc_fname_lists.append([str(x) for x in gc_fnames])
    dirnames = used_dirnames
    objps = np.zeros((chess_shape[0]*chess_shape[1], 3), np.float32)
    objps[:, :2] = chess_block_size * np.mgrid[0:chess_shape[0], 0:chess_shape[1]].T.reshape(-1, 2)
    graycode = GrayCode()
    patterns = graycode.encode((proj_shape[1], proj_shape[0]))
    cam_shape = load_image(gc_fname_lists[0][0], as_grayscale=True).shape
    patch_size_half = int(np.ceil(cam_shape[1] / 180))
    cam_corners_list = []
    cam_objps_list = []
    cam_corners_list2 = []
    proj_objps_list = []
    proj_corners_list = []
    for dname, gc_filenames in zip(dirnames, gc_fname_lists):
        if len(gc_filenames) != len(patterns):
            raise ValueError("invalid number of images in " + dname)
        imgs = load_images(gc_filenames, as_grayscale=True)[..., None]
        forwardmap, fg = graycode.decode(imgs, (proj_shape[1], proj_shape[0]), mode="ij",
                                         bg_threshold=bg_threshold,
                                         output_dir=output_dir, debug=debug)
        black_img = imgs[-1]
        white_img = imgs[-2]
        imgs = imgs[:-2]
        res, cam_corners = cv2.findChessboardCorners(white_img, chess_shape)
        if not res:
            raise RuntimeError("chessboard was not found in {}".format(gc_filenames[-2]))
        cam_objps_list.append(objps)
        cam_corners_list.append(cam_corners)
        proj_objps = []
        proj_corners = []
        cam_corners2 = []
        for corner, objp in zip(cam_corners, objps):
            c_x = int(round(corner[0][0]))
            c_y = int(round(corner[0][1]))
            src_points = []
            dst_points = []
            # todo: vectorize these loops
            for dx in range(-patch_size_half, patch_size_half + 1):
                for dy in range(-patch_size_half, patch_size_half + 1):
                    x = c_x + dx
                    y = c_y + dy
                    # if int(white_img[y, x]) - int(black_img[y, x]) <= black_thr:
                    #     continue
                    if fg[y, x]:
                        proj_pix = forwardmap[y, x]
                        src_points.append((x, y))
                        dst_points.append(gc_step*np.array(proj_pix))
            if len(src_points) < patch_size_half**2:
                if verbose:
                    print('corner {}, {} was skiped because too few decoded pixels found (check your images and thresholds)'.format(c_x, c_y))
                continue
            h_mat, inliers = cv2.findHomography(
                np.array(src_points), np.array(dst_points))
            point = h_mat@np.array([corner[0][0], corner[0][1], 1]).transpose()
            point_pix = point[0:2]/point[2]
            proj_objps.append(objp)
            proj_corners.append([point_pix])
            cam_corners2.append(corner)
        if len(proj_corners) < 3:
            raise RuntimeError("too few corners were found in {} (less than 3)".format(dname))
        proj_objps_list.append(np.float32(proj_objps))
        proj_corners_list.append(np.float32(proj_corners))
        cam_corners_list2.append(np.float32(cam_corners2))

    # Initial solution of camera's intrinsic parameters
    ret, cam_int, cam_dist, cam_rvecs, cam_tvecs = cv2.calibrateCamera(
        cam_objps_list, cam_corners_list, cam_shape, None, None, None, None)
    if verbose:
        print('Initial camera intrinsic parameters: {}'.format(cam_int))
        print('Initial camera distortion parameters: {}'.format(cam_dist))
        print('Initial camera RMS: {}'.format(ret))

    # Initial solution of projector's parameters
    ret, proj_int, proj_dist, proj_rvecs, proj_tvecs = cv2.calibrateCamera(
        proj_objps_list, proj_corners_list, proj_shape, None, None, None, None)
    if verbose:
        print('Initial projector intrinsic parameters: {}'.format(proj_int))
        print('Initial projector distortion parameters: {}'.format(proj_dist))
        print('Initial projector RMS: {}'.format(ret))

    # Stereo calibration for final solution
    ret, cam_int, cam_dist, proj_int, proj_dist, cam_proj_rmat, cam_proj_tvec, E, F = cv2.stereoCalibrate(
        proj_objps_list, cam_corners_list2, proj_corners_list, cam_int, cam_dist, proj_int, proj_dist, None)
    
    if verbose:
        print('RMS: {}'.format(ret))
        print('Camera intrinsic parameters: {}'.format(cam_int))
        print('Camera distortion parameters: {}'.format(cam_dist))
        print('Projector intrinsic parameters: {}'.format(proj_int))
        print('Projector distortion parameters: {}'.format(proj_dist))
        print('Rotation matrix / translation vector from camera to projector (cam2proj transform): {}, {}'.format(cam_proj_rmat, cam_proj_tvec))
    return cam_int, cam_dist, proj_int, proj_dist, cam_proj_rmat, cam_proj_tvec

class GrayCode:
    """
    a class that handles encoding and decoding graycode patterns
    """
    def encode1d(self, length):
        total_images = len(bin(length-1)[2:])

        def xn_to_gray(n, x):
            # infer a coordinate gray code from its position x and index n (the index of the image out of total_images)
            # gray code is obtained by xoring the bits of x with itself shifted, and selecting the n-th bit
            return (x^(x>>1))&(1<<(total_images-1-n))!=0
        
        imgs_code = 255*np.fromfunction(xn_to_gray, (total_images, length), dtype=int).astype(np.uint8)
        return imgs_code
    
    def encode(self, proj_wh, flipped_patterns=True):
        """
        encode projector's width and height into gray code patterns
        :param proj_wh: projector's (width, height) in pixels as a tuple
        :param flipped_patterns: if True, flipped patterns are also generated for better binarization
        :return: a 3D numpy array of shape (total_images, height, width) where total_images is the number of gray code patterns
        """
        width, height = proj_wh
        codes_width_1d = self.encode1d(width)[:, None, :]
        codes_width_2d = codes_width_1d.repeat(height, axis=1)
        codes_height_1d = self.encode1d(height)[:, :, None]
        codes_height_2d = codes_height_1d.repeat(width, axis=2)
        
        img_white = np.full((height, width), 255, dtype=np.uint8)[None, ...]
        img_black = np.full((height, width),  0, dtype=np.uint8)[None, ...]
        all_images = np.concatenate((codes_width_2d, codes_height_2d), axis=0)
        if flipped_patterns:
            all_images = np.concatenate((all_images, 255-codes_width_2d), axis=0)
            all_images = np.concatenate((all_images, 255-codes_height_2d), axis=0)
        all_images = np.concatenate((all_images, img_white, img_black), axis=0)
        return all_images[..., None]
    
    def binarize(self, captures, flipped_patterns=True, bg_threshold=10):
        """
        binarize a batch of images
        :param captures: a 4D numpy array of shape (n, height, width, 1) of captured images
        :param flipped_patterns: if true, patterns also contain their flipped version for better binarization
        :param bg_threshold: if the difference between the pixel in the white&black images is greater than this, pixel is foreground
        :return: a 4D numpy binary array for decoding (total_images, height, width, 1) where total_images is the number of gray code patterns
        and a binary foreground mask (height, width, 1)
        """
        if not -255 <= bg_threshold <= 255:
            raise ValueError('bg_threshold must be between -255 and 255')
        patterns, bw = captures[:-2], captures[-2:]
        foreground = bw[0].astype(np.int32) - bw[1].astype(np.int32) > bg_threshold
        if flipped_patterns:
            orig, flipped = patterns[:len(patterns)//2], patterns[len(patterns)//2:]
            # valid = (orig.astype(np.int32) - flipped.astype(np.int32)) > bin_threshold
            binary = orig > flipped
            # foreground = foreground & np.all(valid, axis=0)  # only pixels that are valid in all images are foreground
        else:  # slightly more naive thresholding
            binary = patterns >= 0.5*(bw[1] + bw[0])
        return binary, foreground

    def decode1d(self, gc_imgs):
        # gray code to binary
        n, h, w = gc_imgs.shape
        binary_imgs = gc_imgs.copy()
        for i in range(1, n):  # xor with previous image except MSB
            binary_imgs[i, :, :] = np.bitwise_xor(binary_imgs[i, :, :], binary_imgs[i-1, :, :])
        # decode binary
        cofficient = np.fromfunction(lambda i,y,x: 2**(n-1-i), (n,h,w), dtype=int)
        img_index = np.sum(binary_imgs * cofficient, axis=0)
        return img_index

    def decode(self, captures, proj_wh,
               flipped_patterns=True, bg_threshold=10,
               mode="xy", output_dir=None, debug=False):
        """
        decodes a batch of images encoded with gray code
        :param captures: a 4D numpy array of shape (n, height, width, c) of captured images
        :param flipped_patterns: if true, patterns also contain their flipped version for better binarization
        :param bg_threshold: a threshold used for background detection using the all-white and all-black captures
        :param mode: "xy" or "ij" decides the order of last dimension coordinates i nthe output (ij -> height first, xy -> width first)
        :return: a 2D numpy array of shape (height, width, 2) specifying the coordinates of decoded result, and foreground mask (height, width)
        """
        if captures.ndim != 4:
            raise ValueError("captures must be a 4D numpy array")
        if captures.dtype != np.uint8:
            raise ValueError("captures must be uint8")
        if output_dir is not None:
            output_dir = Path(output_dir)
            output_dir.mkdir(parents=True, exist_ok=True)
        b, _, _, c = captures.shape
        b = b - 2 # dont count white and black images
        if flipped_patterns:
            b = b // 2  # dont count flipped patterns
        encoded = self.encode(proj_wh, flipped_patterns)  # sanity: encode with same arguments to verify enough captures are present
        if len(encoded) != len(captures):
            raise ValueError("captures must have length of {}".format(len(encoded)))
        if c != 1:  # naively convert to grayscale
            captures = captures.mean(axis=-1, keepdims=True).round().astype(np.uint8)
        imgs_binary, fg = self.binarize(captures, flipped_patterns, bg_threshold)
        imgs_binary = imgs_binary[:, :, :, 0]
        fg = fg[:, :, 0]
        x = self.decode1d(imgs_binary[:b // 2])
        y = self.decode1d(imgs_binary[b // 2:])
        if mode == "ij":
            forward_map = np.concatenate((y[..., None], x[..., None]), axis=-1)
        elif mode == "xy":
            forward_map = np.concatenate((x[..., None], y[..., None]), axis=-1)
        else:
            raise ValueError("mode must be 'ij' or 'xy'")
        if output_dir is not None:
            np.save(Path(output_dir, "forward_map.npy"), forward_map)
            if debug:
                save_images(imgs_binary[..., None], Path(output_dir, "imgs_binary"))
                save_image(fg, Path(output_dir, "foreground.png"))
                composed = forward_map * fg[..., None]
                composed_normalized = composed / np.array([proj_wh[1], proj_wh[0]])
                composed_normalized_8b = to_8b(composed_normalized)
                composed_normalized_8b_3c = np.concatenate((composed_normalized_8b, np.zeros_like(composed_normalized_8b[..., :1])), axis=-1)
                save_image(composed_normalized_8b_3c, Path(output_dir, "forward_map.png"))
        return forward_map, fg
    


import torch
import numpy as np
from PIL import Image

def is_np(x):
    """
    checks if x is a numpy array or torch tensor (will raise an error if x is neither)
    :param x: object to check
    :return: True if x is a numpy array, False if x is a torch tensor
    """
    if type(x) == np.ndarray:
        return True
    elif type(x) == torch.Tensor:
        return False
    else:
        raise ValueError("input must be torch.Tensor or np.ndarray")

def to_hom(x):
    """
    converts a vector to homogeneous coordinates
    :param x: nxc numpy array
    :return: nxc+1 numpy array
    """
    if is_np(x):
        if x.ndim == 1:
            return np.concatenate((x, np.array([1], dtype=x.dtype)))
        else:
            return np.concatenate((x, np.ones((x.shape[0], 1), dtype=x.dtype)), axis=-1)
    else:
        if x.ndim == 1:
            return torch.cat((x, torch.ones(1, device=x.device)))
        else:
            return torch.cat((x, torch.ones(x.shape[0], 1, device=x.device)), dim=-1)

def homogenize(x, keepdim=False):
    """
    normalizes a homogeneous vector by dividing by the last coordinate
    :param x: nx3 numpy array
    :return: nx4 numpy array
    """
    x = (x / x[..., -1:])
    if not keepdim:
        x = x[..., :-1]
    return x

def normalize(x, eps=1e-7):

    if is_np(x):
        return x / (np.linalg.norm(x, axis=-1, keepdims=True) + eps)
    else:
        return x / (torch.norm(x, dim=-1, keepdim=True) + eps)


def broadcast_batch(*args):
    """
    broadcast a list of arrays to the same shape on the batch dimension
    assumes first dimension is batch unless ndim = 1 for all inputs (but then does not broadcast)
    :param args: list of arrays
    :return: list of arrays with the same shape
    """
    shapes = [a.shape for a in args]
    ndims = np.array([a.ndim for a in args])
    new_args = []
    if (ndims <= 1).all():
        for a in args:
            new_args.append(a[None, :])
        return new_args
    elif ndims.min() == 1:
        raise ValueError("cannot broadcast 1d and nd arrays")
    else:
        batch_dim = np.array([s[0] for s in shapes]).max()
        for i, a in enumerate(args):
            if a.shape[0] == batch_dim:
                new_args.append(a)
            else:
                new_args.append(np.broadcast_to(a, (batch_dim, *a.shape[1:])))
    return new_args

def compose_rt(R: np.array, t: np.array, square=False):
    """
    composes a n x 3 x 4 numpy array from rotation and translation.
    will broadcast upon batch dimension if necessary.
    :param R: nx3x3 numpy array
    :param t: nx3 numpy array
    :param square: if True, output will be 4x4, otherwise 3x4
    :return: n x 3 x 4 composition of the rotation and translation
    """
    RR, tt = broadcast_batch(R, t)
    Rt = np.concatenate((RR, tt[:, :, None]), axis=-1)
    if square:
        Rt = to_44(Rt)
    return Rt

def to_44(mat):
    """
    converts a 3x4 to a 4x4 matrix by concatenating 0 0 0 1
    :param mat: dimsx3x4 numpy array (dims can be any number of dims including 0)
    :return: dimsx4x4 numpy array
    """
    if mat.shape[-2:] == (4, 4):
        return mat
    if mat.shape[-2:] != (3, 4):
        raise ValueError("mat must be 3x4")
    if is_np(mat):
        to_cat = np.broadcast_to(np.array([0, 0, 0, 1]), (*mat.shape[:-2], 1, 4))
        new_mat = np.concatenate((mat, to_cat), axis=-2)
    else:
        to_cat = torch.zeros((*mat.shape[:-2], 1, 4), dtype=mat.dtype, device=mat.device)
        to_cat[..., -1] = 1
        new_mat = torch.cat((mat, to_cat), dim=-2)
    return new_mat

def to_34(mat: np.array):
    """
    converts a 4x4 to a 3x4 matrix by removeing the last row
    :param mat: 4x4 numpy array
    :return: 3x4 numpy array
    """
    if mat.ndim == 3:
        if mat.shape[1:] != (4, 4):
            raise ValueError("mat must be 4x4")
        return mat[:, :-1, :]
    else:
        if mat.shape != (4, 4):
            raise ValueError("mat must be 4x4")
        return mat[:-1, :]

def look_at_np(eye, at, up, opengl=False):
    """
    returns a batch of look_at transforms 4x4 (camera->world, the inverse of a ModelView matrix)
    will broadcast upon batch dimension if necessary.
    :param eye: n x 3 from vectors in world space
    :param at: n x 3 at vector in world space
    :param up: n x 3 up vector in world space
    :param opengl: if True, output will be in OpenGL coordinates (z backward, y up) otherwise (z forward, y down)
    :return: n x 4 x 4 transformation matrices (camera->world, the inverse of a ModelView matrix)
    """
    eye, at, up = broadcast_batch(eye, at, up)
    forward = at - eye
    forward = forward / np.linalg.norm(forward, axis=-1, keepdims=True)
    right = np.cross(forward, up)
    right = right / np.linalg.norm(right, axis=-1, keepdims=True)
    up = np.cross(forward, right)
    up = up / np.linalg.norm(up, axis=-1, keepdims=True)
    rot = np.concatenate((right[..., None], up[..., None], forward[..., None]), axis=-1)
    c2w = compose_rt(rot, eye, square=True)
    if opengl:
        c2w = opencv_c2w_to_opengl_c2w(c2w)
    return c2w

def look_at_torch(
        eye:torch.Tensor, #3
        at:torch.Tensor, #3
        up:torch.Tensor, #3
        opengl:bool=False
    ) -> torch.Tensor: #4,4
    """
    creates a lookat transform matrix
    :param eye: where the camera is
    :param at: where the camera is looking
    :param up: the up vector of world space
    :param opengl: if True, output will be in OpenGL coordinates (z backward, y up) otherwise (z forward, y down)
    :return: 4x4 lookat transform matrix
    """
    # todo batch support
    z = (at - eye).type(torch.float32)
    z /= torch.norm(z)
    x = torch.cross(z, up).type(torch.float32)
    x /= torch.norm(x)
    y = torch.cross(z, x).type(torch.float32)
    y /= torch.norm(y)
    c2w = torch.eye(4, device=z.device)
    c2w[:3,:3] = torch.stack([x,y,z],dim=1)
    c2w[:3,3] = eye
    if opengl:
        c2w = opencv_c2w_to_opengl_c2w(c2w)
    return c2w

def orthographic_projection(l, r, b, t, n, f):
    """
    creates an orthographic projection matrix (OpenGL convention)
    :param l: left
    :param r: right
    :param b: bottom
    :param t: top
    :param n: near
    :param f: far
    :return: 4x4 orthographic projection matrix
    """
    dx = r - l
    dy = t - b
    dz = f - n
    rx = -(r + l) / (r - l)
    ry = -(t + b) / (t - b)
    rz = -(f + n) / (f - n)
    return np.array([[2.0/dx,0,0,rx],
                      [0,2.0/dy,0,ry],
                      [0,0,-2.0/dz,rz],
                      [0,0,0,1]])

def perspective_projection(fovy=45, aspect=1.0, n=0.1, f=100.0):
    """
    creates a perspective projection matrix (OpenGL convention)
    :param fovy: field of view in y direction
    :param aspect: aspect ratio
    :param n: near plane
    :param f: far plane
    :return: 4x4 projection matrix
    """
    s = 1.0/np.tan(np.deg2rad(fovy)/2.0)
    sx, sy = s / aspect, s
    zz = (f+n)/(n-f)
    zw = 2*f*n/(n-f)
    return np.array([[sx,0,0,0],
                      [0,sy,0,0],
                      [0,0,zz,zw],
                      [0,0,-1,0]])

def frustum_projection(x0, x1, y0, y1, z0, z1):
    """
    creates a projection matrix from a frustum (OpenGL convention)
    :param x0: left
    :param x1: right
    :param y0: bottom
    :param y1: top
    :param z0: near
    :param z1: far
    :return: 4x4 projection matrix
    """
    a = (x1+x0)/(x1-x0)
    b = (y1+y0)/(y1-y0)
    c = -(z1+z0)/(z1-z0)
    d = -2*z1*z0/(z1-z0)
    sx = 2*z0/(x1-x0)
    sy = 2*z0/(y1-y0)
    return np.array([[sx, 0, a, 0],
                      [ 0,sy, b, 0],
                      [ 0, 0, c, d],
                      [ 0, 0,-1, 0]])

def opengl_project_from_opencv_intrinsics(opencv_intrinsics, width, height, near=0.1, far=100.0):
    """
    given a matrix K from opencv, returns the corresponding projection matrix for OpenGL ("Eye/Camera/View space -> Clip Space")
    :param opencv_intrinsics: 3x3 intrinsics matrix from opencv
    :param width: width of the image
    :param height: height of the image
    :param near: near plane
    :param far: far plane
    :return: 4x4 projection matrix for OpenGL (note: column major)
    """
    fx = opencv_intrinsics[0, 0]
    fy = opencv_intrinsics[1, 1]
    cx = opencv_intrinsics[0, 2]
    cy = opencv_intrinsics[1, 2]
    opengl_mtx = np.array([[2*fx/width, 0.0, (width - 2*cx)/width, 0.0],
                           [0.0, -2*fy/height, (height - 2*cy)/height, 0.0],
                           [0.0, 0.0, (-far - near) / (far - near), -2.0*far*near/(far-near)],
                           [0.0, 0.0, -1.0, 0.0]])
    return opengl_mtx

def opencv_intrinsics_from_opengl_project(opengl_project, width, height):
    """
    given a projection matrix from OpenGL ("Eye/Camera/View space -> Clip Space"), returns a matrix K for opencv
    :param opengl_project: 4x4 projection matrix from OpenGL
    :param width: width of the image
    :param height: height of the image
    :return: 3x3 intrinsics matrix for opencv
    note: assumes camera center is middle of image
    """
    fx = opengl_project[0, 0] * width / 2
    fy = opengl_project[1, 1] * height / 2
    cx = width / 2
    cy = height / 2
    opencv_mtx = np.array([[fx, 0.0, cx],
                           [0.0, fy, cy],
                           [0.0, 0.0, 1]])
    return opencv_mtx

def opengl_c2w_to_opencv_c2w(opengl_transforms):
    """
    given a modelview matrix (World space->Eye/Camera/View space) where z is backward and y is up,
    converts its coordinate system to opencv convention (z forward, y down)
    :param opengl_transforms: 4x4 modelview matrix or batch of 4x4 modelview matrices
    :return: 4x4 modelview matrix in opencv convention
    """
    if opengl_transforms.ndim == 2:
        my_transforms = opengl_transforms[None, ...]
    else:
        my_transforms = opengl_transforms
    if my_transforms.shape[1:] != (4, 4):
        raise ValueError("transform must be 4x4 or batch of 4x4")
    my_transforms[:, :, 1] *= -1
    my_transforms[:, :, 2] *= -1
    return my_transforms.reshape(opengl_transforms.shape)

def opencv_c2w_to_opengl_c2w(opencv_transform):
    """
    converts coordinates of "vision" (opencv) to OpenGL coordinates by flipping y and z axes
    """
    return opengl_c2w_to_opencv_c2w(opencv_transform)

def create_random_cameras_on_unit_sphere(n_cams, radius, normal=None, opengl=False, device="cpu"):
    """
    creates a batch of world2view ("ModelView" matrix) and view2clip ("Projection" matrix) transforms on a unit sphere looking at the center
    :param n_cams: number of cameras
    :param radius: radius of the sphere
    :param normal: if provided, only the hemisphere in the direction of the normal is sampled
    :param opengl: if True, the coordinate system is converted to openGL convention (z backward, y up)
    :param device: device to put the tensors on
    :return: world2view (world2cam), view2clip (requires further processing if opengl=False)
    """
    locs = torch.randn((n_cams, 3), device=device)
    locs = torch.nn.functional.normalize(locs, dim=1, eps=1e-6)
    if normal is not None:
        if normal.ndim == 1:
            normal = normal[None, :]
        normal = torch.nn.functional.normalize(normal, dim=-1, eps=1e-6)
        dot_product = (locs[:, None, :] @ normal[:, :, None]).squeeze()
        locs[dot_product < 0] *= -1
    locs = locs * radius
    matrices = torch.empty((n_cams, 4, 4), dtype=torch.float32, device=device)
    for i in range(len(locs)):
        matrices[i] = look_at_torch(locs[i],
                                    torch.zeros(3, dtype=torch.float32, device=device),
                                    torch.tensor([0.,0.,1.], device=device),
                                    opengl=opengl)
    v2w = matrices  # c2w
    w2v = torch.inverse(v2w)
    v2c = perspective_projection()
    v2c = np.tile(v2c[None, :], (n_cams, 1, 1))
    v2c = torch.tensor(v2c, dtype=torch.float32, device=device)
    return w2v, v2c

def to_np(arr: torch.Tensor):
    """
    converts a tensor to numpy array
    :param arr: tensor
    :return: numpy array
    """
    if type(arr) == torch.Tensor:
        return arr.detach().cpu().numpy()
    elif type(arr) == np.ndarray:
        return arr
    elif type(arr) == Image.Image:
        return np.array(arr)
    return arr.detach().cpu().numpy()

def to_numpy(arr: torch.Tensor):
    """
    converts a tensor to numpy array
    :param arr: tensor
    :return: numpy array
    """
    return to_np(arr)

def to_torch(arr: np.array, device="cpu", dtype=None):
    """
    converts a numpy array to a torch tensor
    :param arr: numpy array
    :param dtype: dtype of the tensor
    :param device: device to put the tensor on
    :return: torch tensor
    """
    if dtype is None:
        return torch.tensor(arr, device=device)
    else:
        return torch.tensor(arr, dtype=dtype, device=device)

def to_8b(x, clip=True):
    """
    convert an array (float, double) array to 8 bit
    :param x: array
    :param clip: if True, clips values to [0,1]
    :return: 8 bit array
    """
    if is_np(x):
        if x.dtype == np.float32 or x.dtype == np.float64:
            if clip:
                x = np.clip(x, 0, 1)
            return (255 * x).round().astype(np.uint8)
        elif x.dtype == bool:
            return x.astype(np.uint8) * 255
        elif x.dtype == np.uint8:
            return x
    else:
        if x.dtype == torch.float32 or x.dtype == torch.float64:
            if clip:
                x = torch.clamp(x, 0, 1)
            return (255 * x).round().type(torch.uint8)
        elif x.dtype == torch.bool:
            return x.type(torch.uint8) * 255
        elif x.dtype == torch.uint8:
            return x

def to_float(x, clip=True):
    """
    convert a 8bit or bool array to float
    :param x: array
    :param clip: if True, clips values to [0,1]
    :return: float array
    """
    if is_np(x):
        if x.dtype == np.uint8:
            return x.astype(np.float32) / 255
        elif x.dtype == np.float32:
            if clip:
                x = np.clip(x, 0, 1)
            return x
        elif x.dtype == bool:
            return x.astype(np.float32)
        else:
            raise ValueError("unsupported dtype")
    else:
        if x.dtype == torch.uint8:
            return x.to(torch.float32) / 255
        elif x.dtype == torch.float32:
            if clip:
                x = torch.clamp(x, 0, 1)
            return x
        elif x.dtype == torch.bool:
            return x.to(torch.float32)
        else:
            raise ValueError("unsupported dtype")

def to_PIL(x: np.array):
    """
    convert a numpy float array to a PIL image
    """
    if x.ndim == 3:
        return Image.fromarray(to_8b(x))
    elif x.ndim == 2:
        return Image.fromarray(to_8b(x[:, :, None]), mode="L")
    else:
        raise ValueError("unsupported array dimensions")

def map_range(x, out_min, out_max):
    """
    given an input and a range, maps it to a new range
    """
    if type(x) == np.ndarray:
        return np.clip((x-x.min()) * (out_max-out_min) / (x.max()-x.min()) + out_min, out_min, out_max)
    elif type(x) == torch.Tensor:
        return torch.clamp((x-x.min()) * (out_max-out_min) / (x.max()-x.min()) + out_min, out_min, out_max)
    else:
        raise ValueError("unsupported type")

def map_to_01(x):
    return map_range(x, 0, 1)


def vec2skew(v):
    """
    returns the skew operator matrix given a vector
    :param v:  (3, ) torch tensor
    :return:   (3, 3)
    """
    return batch_vec2skew(v[None, :])[0]

def batch_vec2skew(v):
    """
    returns a batch of skew operator matrices given a batch of vectors
    :param v:  (B, 3) torch tensor
    :return:   (B, 3, 3)
    """
    zero = torch.zeros_like(v[:, 0:1])
    skew_v0 = torch.cat([zero, -v[:, 2:3], v[:, 1:2]], dim=1)  # (B, 3)
    skew_v1 = torch.cat([v[:, 2:3], zero, -v[:, 0:1]], dim=1)
    skew_v2 = torch.cat([-v[:, 1:2], v[:, 0:1], zero], dim=1)
    skew_v = torch.stack([skew_v0, skew_v1, skew_v2], dim=1)  # (B, 3, 3)
    return skew_v  # (B, 3, 3)

def batch_rotvec2mat(r: torch.Tensor):
    """so(3) vector to SO(3) matrix
    :param r: (3, ) axis-angle, torch tensor
    :return:  (3, 3)
    """
    if r.ndim != 2:
        raise ValueError("r must be 2D tensor.")
    skew_r = batch_vec2skew(r)  # (3, 3)
    norm_r = r.norm(dim=-1, keepdim=True)[:, :, None] + 1e-15
    eye = torch.eye(3, dtype=torch.float32, device=r.device)[None, :]
    R = eye + (torch.sin(norm_r) / norm_r) * skew_r + ((1 - torch.cos(norm_r)) / norm_r ** 2) * (skew_r @ skew_r)
    return R

def rotvec2mat(r: torch.Tensor):
    """so(3) vector to SO(3) matrix
    :param r: (3, ) axis-angle, torch tensor
    :return:  (3, 3)
    """
    return batch_rotvec2mat(r[None, :])[0]

def mat2rotvec(r: torch.Tensor):
    """SO(3) matrix to so(3) vector
    :param r: (3, ) axis-angle, torch tensor
    :return:  (3, 3)
    """
    e, v = torch.linalg.eig(r)
    rotvec = v.real[:, torch.isclose(e.real, torch.ones(1))].squeeze()
    up = torch.tensor([0, 1, 0], dtype=torch.float)
    # ortho = torch.cross(rotvec, up)
    # ortho = ortho / ortho.norm(keepdim=True)
    # angle1 = torch.arccos(ortho.dot(r @ ortho))
    angle2 = torch.arccos((torch.trace(r) - 1) / 2)
    raise NotImplementedError("please verify this function implementation before usage")
    return rotvec * angle2

def random_qvec(n: int):
    """
    generate random quaternions representing rotations
    :param n: Number of quaternions in a batch to return.
    :return: Quaternions as tensor of shape (N, 4).
    """
    o = np.random.randn(n, 4)
    s = (o * o).sum(1)
    denom = np.copysign(np.sqrt(s), o[:, 0])[:, None]
    o = o / denom
    return o

def random_affine(ang_range=20.0, trans_range=10.0, scale=2.0):
    """
    generates a random 2D affine transformation matrix
    """
    R = np.eye(3)
    ang_rot = np.random.uniform(ang_range)-ang_range/2
    tx = trans_range*np.random.uniform()-trans_range/2
    ty = trans_range*np.random.uniform()-trans_range/2
    sx = np.random.rand() * scale
    sy = np.random.rand() * scale
    R[0, 0] = np.cos(ang_rot * np.pi / 180)
    R[0, 1] = -np.sin(ang_rot * np.pi / 180)
    R[1, 0] = np.sin(ang_rot * np.pi / 180)
    R[1, 1] = np.cos(ang_rot * np.pi / 180)
    T = np.eye(3)
    T[0, 2] = tx
    T[1, 2] = ty
    S = np.eye(3)
    S[0, 0] = sx
    S[1, 1] = sy
    H = T @ S @ R
    return H

def random_perspective():
    """
    generates a random 2D perspective transformation matrix
    """
    P = random_affine()
    P[2, 0] = np.random.rand() / 100
    P[2, 1] = np.random.rand() / 100
    return P

def random_vectors_on_sphere(n, normal=None, device="cpu"):
    """
    create a batch of uniformly distributed random unit vectors on a sphere
    note: if normal is provided, returns random unit vectors on the hemisphere around the normal, but isn't uniform anymore.
    :param n: number of vectors
    :param normal: normals to orient the hemisphere (,3) or (n,3)
    :param device: device to put the tensors on
    :return: tensor of shape (n, 3)
    """
    locs = torch.randn((n, 3), device=device)
    locs = torch.nn.functional.normalize(locs, dim=1, eps=1e-6)
    if normal is not None:
        if normal.ndim == 1:
            normal = normal[None, :]
        normal = torch.nn.functional.normalize(normal, dim=-1, eps=1e-6)
        dot_product = (locs[:, None, :] @ normal[:, :, None]).squeeze()
        locs[dot_product < 0] *= -1
    return locs

def qvec2mat(qvec):
    """
    Converts a quaternion to a rotation matrix
    :param qvec: tensor of size 4 where real part is first (a + bi + cj + dk) => (a, b, c, d)
    :return: rotation matrix 3x3
    """
    return batch_qvec2mat(qvec[None, :])[0]

def batch_qvec2mat(qvecs):
    """
    Converts a batch of quaternions to a batch of rotation matrices
    :param qvec: tensor of size nx4 where real part is first (a + bi + cj + dk) => (a, b, c, d)
    :return: rotation matrix nx3x3
    """
    if qvecs.shape[-1] != 4:
        raise ValueError("quaternions must be of shape (..., 4)")
    if qvecs.ndim == 1:
        qvecs = qvecs[None, :]
    if qvecs.ndim > 2:
        raise ValueError("quaternions must be of shape (..., 4)")
    if is_np(qvecs):
        r, i, j, k = [x[0] for x in np.split(qvecs, 4, -1)]
        stacking_func = np.stack
    else:
        r, i, j, k = torch.unbind(qvecs, -1)
        stacking_func = torch.stack
    two_s = 2.0 / (qvecs * qvecs).sum(-1)
    o = stacking_func([1 - two_s * (j * j + k * k),
                       two_s * (i * j - k * r),
                       two_s * (i * k + j * r),
                       two_s * (i * j + k * r),
                       1 - two_s * (i * i + k * k),
                       two_s * (j * k - i * r),
                       two_s * (i * k - j * r),
                       two_s * (j * k + i * r),
                       1 - two_s * (i * i + j * j)],
                       -1)
    return o.reshape(qvecs.shape[:-1] + (3, 3))

def mat2qvec_numpy(R: np.array):
    """
    converts a rotation matrix to a quaternion
    :param R: np array of size 3x3
    :return: qvec (4,) xyzw
    """
    q = np.empty((R.shape[0], 4), dtype=R.dtype)
    trace = np.trace(R)
    expanded_trace = np.concatenate((np.diag(R), trace), axis=0)
    choice = np.argmax(expanded_trace, axis=-1)
    mask = choice != 3
    i = choice[mask]
    j = (i+1) % 3
    k = (j+1) % 3
    ii = np.concatenate((i, i), axis=1)
    ij = np.concatenate((i, j), axis=1)
    ik = np.concatenate((i, k), axis=1)
    jk = np.concatenate((j, k), axis=1)
    q[:, 0] = R[:, 2, 1] - R[:, 1, 2]
    q[:, 1] = R[:, 0, 2] - R[:, 2, 0]
    q[:, 2] = R[:, 1, 0] - R[:, 0, 1]
    q[:, 3] = 1 + trace
    q[mask, 0] = 1 - trace + 2*R[ii]
    q[mask, 1] = R[np.flip(ij)] + R[ij]
    q[mask, 2] = R[np.flip(ik)] + R[ik]
    q[mask, 3] = R[np.flip(jk)] - R[jk]
    raise NotImplementedError("please verify this function implementation before usage")
    # Rxx, Ryx, Rzx, Rxy, Ryy, Rzy, Rxz, Ryz, Rzz = R.flat
    # K = np.array([
    #     [Rxx - Ryy - Rzz, 0, 0, 0],
    #     [Ryx + Rxy, Ryy - Rxx - Rzz, 0, 0],
    #     [Rzx + Rxz, Rzy + Ryz, Rzz - Rxx - Ryy, 0],
    #     [Ryz - Rzy, Rzx - Rxz, Rxy - Ryx, Rxx + Ryy + Rzz]]) / 3.0
    # eigvals, eigvecs = np.linalg.eigh(K)
    # qvec = eigvecs[[3, 0, 1, 2], np.argmax(eigvals)]
    # if qvec[0] < 0:
    #     qvec *= -1
    # return qvec
    return q

def mat2qvec(matrix):
    return batch_mat2qvec(matrix[None, :])[0]

def batch_mat2qvec(matrix):
    """
    Converts a batch of rotation matrices to a batch of quaternions
    :param matrix: tensor of size nx3x3
    :return: qvec: tensor of size nx4 where real part is first (a + bi + cj + dk) => (a, b, c, d)
    implementation is exact copy of pytorch3d implementation
    """
    if matrix.size(-1) != 3 or matrix.size(-2) != 3:
        raise ValueError(f"Invalid rotation matrix shape {matrix.shape}.")
    batch_dim = matrix.shape[:-2]
    m00, m01, m02, m10, m11, m12, m20, m21, m22 = torch.unbind(
        matrix.reshape(batch_dim + (9,)), dim=-1
    )
    q = torch.stack(
            [
                1.0 + m00 + m11 + m22,
                1.0 + m00 - m11 - m22,
                1.0 - m00 + m11 - m22,
                1.0 - m00 - m11 + m22,
            ],
            dim=-1,
        )
    q_abs = torch.zeros_like(q)
    positive_mask = q > 0
    q_abs[positive_mask] = torch.sqrt(q[positive_mask])
    # we produce the desired quaternion multiplied by each of r, i, j, k
    quat_by_rijk = torch.stack(
        [
            torch.stack([q_abs[..., 0] ** 2, m21 - m12, m02 - m20, m10 - m01], dim=-1),
            torch.stack([m21 - m12, q_abs[..., 1] ** 2, m10 + m01, m02 + m20], dim=-1),
            torch.stack([m02 - m20, m10 + m01, q_abs[..., 2] ** 2, m12 + m21], dim=-1),
            torch.stack([m10 - m01, m20 + m02, m21 + m12, q_abs[..., 3] ** 2], dim=-1),
        ],
        dim=-2,
    )
    # We floor here at 0.1 but the exact level is not important; if q_abs is small,
    # the candidate won't be picked.
    flr = torch.tensor(0.1).to(dtype=q_abs.dtype, device=q_abs.device)
    quat_candidates = quat_by_rijk / (2.0 * q_abs[..., None].max(flr))
    # if not for numerical problems, quat_candidates[i] should be same (up to a sign),
    # forall i; we pick the best-conditioned one (with the largest denominator)
    return quat_candidates[torch.nn.functional.one_hot(q_abs.argmax(dim=-1), num_classes=4) > 0.5, :].reshape(batch_dim + (4,))


import numpy as np
import torch
from PIL import Image, ImageDraw, ImageFont
from scipy import interpolate, spatial
from .core import to_8b, to_float, to_hom, homogenize, broadcast_batch, is_np, to_torch, to_np
from .structures import get_gizmo_coords
from .gsoup_io import save_image

def alpha_compose(images, backgrounds=None, bg_color=None):
    """
    composes a single or batch of RGBA images into a single or batch of RGB images.
    if no backgrounds or bg_color is provided, the background is assumed to be black.
    :param image: b x H x W x 4 or H x W x 4
    :param background: b x H x W x 3 or H x W x 3
    :param bg_color: 3 or b x 3 float32 array
    :return: b x H x W x 3 or H x W x 3
    """
    if images.ndim != 3 and images.ndim != 4:
        raise ValueError("image must be 3 or 4 dimensional")
    if images.shape[-1] != 4:
        raise ValueError("image must have 4 channels")
    if backgrounds is not None:
        if images.shape[:-1] != backgrounds.shape[:-1]:
            raise ValueError("backgrounds must have same shape as images")
    if is_np(images):
        if bg_color is None:
            bg_color = np.array([0., 0., 0.]).astype(np.float32)
        if images.dtype != np.float32:
            images = to_float(images)
    else:
        if bg_color is None:
            bg_color = torch.tensor([0., 0., 0.], dtype=images.dtype, device=images.device)
        if images.dtype != torch.float32:
            images = to_float(images)
    if backgrounds is not None:
        if backgrounds.dtype != np.float32:
            backgrounds = to_float(backgrounds)
        bg_color = backgrounds
    alpha = images[..., 3:4]
    rgb = images[..., :3]
    return alpha * rgb + (1 - alpha) * bg_color

def draw_text_on_image(images, text_per_image, fill_white=True):
    """
    writes text on images given as np array (b x H x W x 3)
    :param images: (b x H x W x 3) numpy array
    :param text_per_image: b x 1 numpy array of strings
    :param fill_white: if True, text is white, otherwise black
    :return: new (b x H x W x 3) numpy array with text written
    """
    is_numpy = is_np(images)
    if not is_numpy:
        device = images.device
        images = to_np(images)
    is_float = images.dtype == np.float32
    if is_float:
        images = to_8b(images)
    rgbs = [Image.fromarray(x) for x in images]
    font = ImageFont.truetype("./FreeMono.ttf", 48)
    if fill_white:
        fill = "white"
    else:
        fill = "black"
    for i, rgb in enumerate(rgbs):
        text = text_per_image[i]
        ImageDraw.Draw(rgb).text((0, 0), text, fill=fill, font=font)
    rgbs = np.array([np.asarray(rgb) for rgb in rgbs])
    if is_float:
        rgbs = to_float(rgbs)
    if not is_numpy:
        rgbs = to_torch(rgbs, device=device)
    return rgbs

def draw_gizmo_on_image(np_images, w2c, opengl=False, scale=.05):
    """
    adds a gizmo to a batch of np images.
    note: will broadcast np_images and w2c against eachother.
    :param np_images: b x H x W x 3
    :param w2c: b x 3 x 4 w2c transforms (opencv conventions)
    :param opengl: if True, the w2c transforms are assumed to be in OpenGL conventions, else OpenCV conventions
    for opengl, w2c should be a bx4x4 matrix converting from world to *CLIP* space.
    :param scale: scale of the gizmo
    :return: b x H x W x 3
    """
    new_images = []
    if np_images.ndim != 4:
        raise ValueError("np_images must be b x H x W x 3")
    if w2c.ndim != 3:
        raise ValueError("KRt must be b x 3 x 4")
    np_images, w2c = broadcast_batch(np_images, w2c)
    for i, np_image in enumerate(np_images):
        pil_image = Image.fromarray(to_8b(np_image))
        W, H = pil_image.size
        gizmo_cords, _, _ = get_gizmo_coords(scale)
        gizmo_hom = to_hom(gizmo_cords)
        verts_clip = (w2c[i] @ gizmo_hom.T).T
        verts_clip = homogenize(verts_clip)
        if opengl:
            verts_clip = verts_clip[:, :2]
            verts_screen = np.array([W, H]) * (verts_clip + 1) / 2
            verts_screen[:, 1] *= -1
        else:
            verts_screen = verts_clip
        desired_loc = np.array([W - 40, H - 40])
        verts_screen += desired_loc - verts_screen[0]
        draw = ImageDraw.Draw(pil_image)
        draw.line((tuple(verts_screen[0]), tuple(verts_screen[1])), fill="red", width = 0)
        draw.line((tuple(verts_screen[0]), tuple(verts_screen[2])), fill="green", width = 0)
        draw.line((tuple(verts_screen[0]), tuple(verts_screen[3])), fill="blue", width = 0)
        new_images.append(np.array(pil_image))
    return (np.array(new_images) / 255.).astype(np.float32)

def merge_figures_with_line(img1, img2, lower_intersection=0.6, angle=np.pi/2, line_width=5, line_color=[255, 255, 255, 255]):
    """
    merges two np images (H x W x 3) with a white line in between
    :param img1: (H x W x 3) numpy array
    :param img2: (H x W x 3) numpy array
    :param lower_intersection: lower intersection of the line with the images
    :param angle: angle of the line
    :param line_width: width of the line
    :param line_color: color of the line
    :return: new (H x W x 3) numpy array with line in between
    """
    if img1.dtype != np.uint8:
        line_color = np.array(line_color) / 255
    else:
        line_color = np.array(line_color, dtype=np.uint8)
    combined = np.ascontiguousarray(np.broadcast_to(line_color, img1.shape))
    y, x, _ = img1.shape
    yy, xx = np.mgrid[:y, :x]
    img1_positions = (xx-lower_intersection*x)*np.tan(angle)-line_width//2>(yy-y)
    img2_positions = (xx-lower_intersection*x)*np.tan(angle)+line_width//2<(yy-y)
    combined[img1_positions] = img1[img1_positions]
    combined[img2_positions] = img2[img2_positions]
    return combined

def generate_voronoi_diagram(height, width, num_cells=1000, bg_color="white", dst=None):
    """
    generate a voronoi diagram HxWx3 with random colored cells
    :param height: height of the image
    :param width: width of the image
    :param num_cells: number of cells
    :param bg_color: background color
    :param dst: if not None, the image is written to this path
    :return: (H x W x 3) numpy array
    """
    nx = np.random.rand(num_cells) * width
    ny = np.random.rand(num_cells) * height
    nxy = np.stack((nx, ny), axis=-1)
    img = Image.new("RGB", (width, height), bg_color)
    vor = spatial.Voronoi(nxy)
    polys = vor.regions
    vertices = vor.vertices
    for poly in polys:
        polygon = vertices[poly]
        if len(poly) > 0 and np.all(np.array(poly) > 0):
            img1 = ImageDraw.Draw(img)
            img1.polygon(list(map(tuple, polygon)), fill=tuple(np.random.randint(0, 255, size=(3,))))
    if dst is not None:
        img.save(str(dst))
    return np.array(img)

def generate_dot_pattern(height, width, background="black", radius=5, spacing=50, dst=None):
    """
    generates an image with colored circles in a grid
    :param height: height of the image
    :param width: width of the image
    :param background: background color
    :param radius: radius of the circles in pixels
    :param spacing: spacing between the circles in pixels
    :param dst: if not None, the image is written to this path
    :return: (H x W x 3) numpy array (uint8)
    """
    img = Image.new("RGB", (width, height), background)
    img1 = ImageDraw.Draw(img)
    for i in range(0, width, spacing):
        for j in range(0, height, spacing):
            img1.ellipse([i, j, i + radius*2, j + radius*2], fill=tuple(np.random.randint(0, 255, size=(3,))))
    if dst is not None:
        img.save(str(dst))
    return np.array(img)

def generate_checkerboard(h, w, blocksize):
    """
    generates a checkerboard pattern
    note: if blocksize is not a divisor of w or h, the pattern will have extra "crops" at the edges
    :param h: height of the image
    :param w: width of the image
    :param blocksize: size of the squares
    :return: (H x W x 1) numpy array (bool)
    """
    c0, c1 = 0, 1  # color of the squares, for binary these are just 0,1
    tile = np.array([[c0, c1],[c1, c0]], dtype=bool).repeat(blocksize, axis=0).repeat(blocksize, axis=1)[..., None]
    grid = np.tile(tile, ( h//(2*blocksize)+1, w//(2*blocksize)+1, 1))
    return grid[:h,:w]

def generate_stripe_pattern(height, width, background="black", direction="vert", thickness=5, spacing=50, dst=None):
    """
    generates an image with colored stripes in a certain direction
    :param height: height of the image
    :param width: width of the image
    :param background: background color
    :param direction: direction of the stripes ("vert", "hor", "both")
    :param thickness: thickness of the stripes
    :param spacing: spacing between the stripes
    :param dst: if not None, the image is written to this path
    :return: (H x W x 3) numpy array (uint8)
    """
    img = Image.new("RGB", (width, height), background)
    img1 = ImageDraw.Draw(img)
    if direction == "vert":
        for i in range(0, width, spacing):
            img1.rectangle([i, 0, i + thickness, height], fill=tuple(np.random.randint(0, 255, size=(3,))))
    elif direction == "hor":
        for i in range(0, height, spacing):
            img1.rectangle([0, i, width, i + thickness], fill=tuple(np.random.randint(0, 255, size=(3,))))
    elif direction == "both":
        for i in range(0, width, spacing):
            img1.rectangle([i, 0, i + thickness, height], fill=tuple(np.random.randint(0, 255, size=(3,))))
        for i in range(0, height, spacing):
            img1.rectangle([0, i, width, i + thickness], fill=tuple(np.random.randint(0, 255, size=(3,))))
    else:
        raise ValueError("direction must be either 'vert', 'hor' or 'both'")
    if dst is not None:
        img.save(str(dst))
    return np.array(img)

def generate_lollipop_pattern(height, width, background="black", n=15, m=8, dst=None):
    """
    generates an image with a lollipop pattern
    :param height: height of the image
    :param width: width of the image
    :param background: background color
    :param n: number of circles in the pattern
    :param m: number of lines in the pattern
    :param dst: if not None, the image is written to this path
    :return: (H x W x 3) numpy array (uint8)
    """
    spacing_x = width // (2*n)
    spacing_y = height // (2*n)
    spacing_angle = 360 // m
    img = Image.new("RGB", (width, height), background)
    img1 = ImageDraw.Draw(img)
    for j in range(m):
        for i in range(n):
            x0 = spacing_x * i
            y0 = spacing_y * i
            start_angle = j * spacing_angle
            end_angle = 360
            img1.pieslice([x0, y0, width - x0, height - y0], start=start_angle, end=end_angle, fill=tuple(np.random.randint(0, 255, size=(3,))))
    if dst is not None:
        img.save(str(dst))
    return np.array(img)

def generate_concentric_circles(height, width, background="black", n=15, dst=None):
    """
    generates an image with colored concentric circles
    :param height: height of the image
    :param width: width of the image
    :param background: background color
    :param n: number of circles to draw
    :param dst: if not None, the image is written to this path
    """
    spacing_x = width // (2*n)
    spacing_y = height // (2*n)
    img = Image.new("RGB", (width, height), background)
    img1 = ImageDraw.Draw(img)
    for i in range(n):
        x0 = spacing_x * i
        y0 = spacing_y * i
        img1.ellipse([x0, y0, width - x0, height - y0], fill=tuple(np.random.randint(0, 255, size=(3,))))
    if dst is not None:
        img.save(str(dst))
    return np.array(img)

def generate_gray_gradient(height, width, grayscale=False, vertical=True, flip=False, bins=10, dst=None):
    """
    generate a gray gradient image HxWx3
    :param height: height of the image
    :param width: width of the image
    :param grayscale: if True, the image is grayscale
    :param vertical: if True, the gradient is vertical
    :param flip: if True, the gradient is flipped
    :param bins: number of bins
    :param dst: if not None, the image is written to this path
    :return: (H x W x 3) numpy array
    """
    bins = np.clip(bins, 1, 256)
    colors = np.linspace(0, 255, num=bins).astype(np.uint8)
    if vertical:
        n_bins = height // bins
        up_to_max_intensity = colors.repeat(n_bins)[:height]
        pad_size = height - up_to_max_intensity.shape[0]
        if pad_size > 0:
            channel = np.concatenate((up_to_max_intensity, np.full(pad_size, 255, dtype=np.uint8)))
        else:
            channel = up_to_max_intensity
        img = channel[:, None].repeat(width, axis=1)
        if flip:
            img = np.flip(img, axis=0)
    else:
        n_bins = width // bins
        up_to_max_intensity = colors.repeat(n_bins)[:width]
        pad_size = width - up_to_max_intensity.shape[0]
        if pad_size > 0:
            channel = np.concatenate((up_to_max_intensity, np.full(pad_size, 255, dtype=np.uint8)))
        else:
            channel = up_to_max_intensity
        img = channel[None, :].repeat(height, axis=0)
        if flip:
            img = np.flip(img, axis=1)
    if not grayscale:
        img = img[:, :, None].repeat(3, axis=-1)
    if dst is not None:
        save_image(img, dst)
    return img

def interpolate_single_channel(image: np.ndarray, mask: np.ndarray, method: str = "linear", fill_value: int = 0):
    """
    :param image: (H x W) numpy array
    :param mask: (H x W) boolean numpy array, True indicates missing values
    :param method: interpolation method, one of
        'nearest', 'linear', 'cubic'.
    :param fill_value: which value to use for filling up data outside the
        convex hull of known pixel values.
        Default is 0, Has no effect for 'nearest'.
    :return: the image with missing values interpolated
    """
    h, w = image.shape[:2]
    xx, yy = np.meshgrid(np.arange(w), np.arange(h))

    known_x = xx[~mask]
    known_y = yy[~mask]
    known_v = image[~mask]
    missing_x = xx[mask]
    missing_y = yy[mask]

    interp_values = interpolate.griddata(
        (known_x, known_y), known_v, (missing_x, missing_y),
        method=method, fill_value=fill_value
    )

    interp_image = image.copy()
    interp_image[missing_y, missing_x] = interp_values

    return interp_image

def interpolate_multi_channel(image: np.ndarray, mask: np.ndarray):
    """
    given a multi channel image and a mask, interpolate the values where mask is true (per channel interpolation)
    :param image: (H x W x C) numpy array
    :param mask: (H x W) numpy array
    :return: (H x W x C) numpy array
    """
    if image.ndim != 3:
        raise ValueError("image must have atleast 1 channel")
    interpolated = np.zeros_like(image)
    for channel in range(image.shape[-1]):
        interpolated_channel = interpolate_single_channel(image[:, :, channel], mask)
        interpolated[:, :, channel] = interpolated_channel
    return interpolated

def image_grid(images, rows, cols):
    """
    :param images: list of images
    :param rows: number of rows
    :param cols: number of cols
    :return: grid image
    """
    if images.ndim != 4:
        raise ValueError("images must be a 4D array")
    if len(images) != rows * cols:
        raise ValueError("number of images must be equal to rows * cols")
    tmp = images.reshape(rows, cols, images.shape[1], images.shape[2], -1)
    if type(tmp) == torch.Tensor:
        result = tmp.permute(0, 2, 1, 3, 4).reshape(rows * images.shape[1], cols * images.shape[2], -1)
    elif type(tmp) == np.ndarray:
        result = tmp.transpose(0, 2, 1, 3, 4).reshape(rows * images.shape[1], cols * images.shape[2], -1)
    return result

def resize_images_naive(images, H, W, channels_last=True, mode="mean"):
    """
    resize images to output_size, but only if the output size has a common divisor of the input size
    :param images: numpy array of images (N x H x W x C)
    :param H: output height size that has a common divisor with the input height size
    :param W: output width size that has a common divisor with the input width size
    :param channels_last: if True, the images are provided in channels last format (and so will the output)
    :param mode: one of "max", "mean"
    :return: np array of resized images
    """
    if images.ndim != 4:
        raise ValueError("images must be a 4D array")
    if images.shape[1] != images.shape[2]:
        raise ValueError("images must be square")
    if not channels_last:
        raise NotImplementedError("only channels last is supported")
    channels_size = images.shape[-1]
    input_size = np.array(images.shape[1:3])
    output_size = np.array([H, W])
    bin_size = input_size // output_size
    if mode == "max":
        small_images = images.reshape((images.shape[0], output_size[0], bin_size[0], output_size[1], bin_size[1], channels_size)).max(4).max(2)
    elif mode == "mean":
        small_images = images.reshape((images.shape[0], output_size[0], bin_size[0], output_size[1], bin_size[1], channels_size)).mean(4).mean(2)
        if images.dtype == np.uint8:
            small_images = small_images.astype(np.uint8)
    else:
        raise ValueError("mode must be one of 'max', 'mean'")
    return small_images

def pad_to_square(images, color=None):
    """
    pads a batch of images to a square shape
    note: smaller dimension is padded
    :param image: numpy image b x h x w x c
    :param color: color to pad with
    :return: padded image
    """
    if images.ndim != 4:
        raise ValueError("image must be a 4D array")
    diff = images.shape[1] - images.shape[2]
    if diff == 0:
        return images
    if diff > 0:
        return pad_to_res(images, images.shape[1], images.shape[1], color)
    else:
        return pad_to_res(images, images.shape[2], images.shape[2], color)

def crop_to_square(images):
    """
    crops a batch of images to a square shape
    note: bigger dimension is cropped
    :param img: numpy image h x w x c
    :return: the cropped square image
    """
    if images.ndim != 4:
        raise ValueError("image must be a 4D array")
    if images.shape[1] > images.shape[2]:
        s = int((images.shape[1] - images.shape[2]) / 2)
        return images[s:(s + images.shape[2])]
    else:
        s = int((images.shape[2] - images.shape[1]) / 2)
        return images[:, :, s:(s + images.shape[1])]

def pad_to_res(images, res_h, res_w, bg_color=None):
    """
    pads a batch of numpy images to a specific resolution
    :param image: numpy image b x h x w x c
    :param res_h: height of the output image
    :param res_w: width of the output image
    :param bg_color: background color c (defaults to black)
    :return: padded image b x res_h x res_w x c
    """
    if bg_color is None:
        if is_np(images):
            bg_color = np.zeros(images.shape[-1], dtype=images.dtype)
        else:
            bg_color = torch.zeros(images.shape[-1], dtype=images.dtype, device=images.device)
    if images.ndim != 4:
        raise ValueError("image must be a 4D array")
    b, h, w, c = images.shape
    if h > res_h or w > res_w:
        raise ValueError("images dimensions is larger than the output resolution")
    if h == res_h and w == res_w:
        return images
    if bg_color.shape[0] != c:
        raise ValueError("background color must have the same number of channels as the image")
    bg_color = bg_color[None, None, None, :]
    if is_np(images):
        output = np.zeros((b, res_h, res_w, c), dtype=images.dtype)
    else:
        output = torch.zeros((b, res_h, res_w, c), dtype=images.dtype, device=images.device)
    output[:, :, :, :] = bg_color
    corner_left = (res_w - w) // 2
    corner_top = (res_h - h) // 2
    output[:, corner_top:corner_top + h, corner_left:corner_left + w, :] = images
    return output

def crop_center(images, dst_h, dst_w):
    """
    crops a batch of images to a specific resolution, but the crop comes from the center of the image
    :param image: numpy (or torch) array b x h x w x c
    :param dst_h: height of the output image
    :param dst_w: width of the output image
    :return: cropped image b x dst_h x dst_w x c
    """
    if images.ndim != 4:
        raise ValueError("image must be a 4D array")
    _, h, w, _ = images.shape
    if h < dst_h or w < dst_w:
        raise ValueError("images dimensions is smaller than the output resolution")
    if h == dst_h and w == dst_w:
        return images
    corner_left = (w - dst_w) // 2
    corner_top = (h - dst_h) // 2
    output = images[:, corner_top:corner_top + dst_h, corner_left:corner_left + dst_w, :]
    return output

def mask_regions(images, start_h, end_h, start_w, end_w):
    """
    masks a batch of numpy image with black background outside of region of interest (roi)
    :param image: numpy image b x h x w x c
    :param start_h: where does the roi height start
    :param end_h: where does the roi height end
    :param start_w: where does the roi width start
    :param end_w: where does the roi width end
    :return: masked image
    """
    if images.ndim != 4:
        raise ValueError("image must be a 4D array")
    _, h, w, _ = images.shape
    if start_h < 0 or start_w < 0 or end_h > h or end_w > w:
        raise ValueError("Values exceed image resolution")
    output = np.zeros_like(images)
    output[:, start_h:end_h, start_w:end_w, :] = images[start_h:end_h, start_w:end_w, :]
    return output

def adjust_contrast_brightness(img, alpha, beta=None):
    """
    adjusts image contrast and brightness using naive gain and bias factors
    :param img: input image numpy array (n x h x w x 3), float values between 0 and 1
    :param alpha: gain factor ("contrast") between 0 and inf
    :param beta: bias factor ("brightness") between -inf and inf (but typically between -1 and 1)
    :return: the new image
    """
    if img.dtype != np.float32:
        raise ValueError("img must be a float32 numpy array (0-1)")
    if beta is None:  # if beta is not provided, set to factor of alpha
        beta = 0.5 - alpha / 2
    new_img = img * alpha + beta
    new_img[new_img < 0] = 0
    new_img[new_img > 1] = 1
    return new_img.astype(np.uint8)

def change_brightness(input_img, brightness=0):
    """
    changes brightness of an image or batch of images
    :param input_img a numpy or torch tensor of float values between 0 and 1 (n x h x w x 3)
    :param brightness a number between -255 to 255 (0=no change)
    :return the new image
    """
    if input_img.dtype != np.float32 and input_img.dtype != torch.float32:
        raise ValueError("input_img must be a float32 array (0-1)")
    if brightness != 0:
        if brightness > 0:
            shadow = brightness
            highlight = 255
        else:
            shadow = 0
            highlight = 255 + brightness
        alpha_b = (highlight - shadow)/255
        gamma_b = shadow / 255
    else:
        alpha_b = 1
        gamma_b = 0
    return input_img*alpha_b + gamma_b

def linear_to_srgb(linear):
    """
    converts linear RGB to sRGB, see https://en.wikipedia.org/wiki/SRGB.
    note: linear is expected to be in the range [0, 1]
    """
    eps = np.finfo(np.float32).eps
    srgb0 = 323 / 25 * linear
    srgb1 = (211 * np.maximum(eps, linear)**(5 / 12) - 11) / 200
    return np.where(linear <= 0.0031308, srgb0, srgb1)

def srgb_to_linear(srgb):
    """
    converts linear RGB to sRGB, see https://en.wikipedia.org/wiki/SRGB.
    note: srgb is expected to be in the range [0, 1]
    """
    eps = np.finfo(np.float32).eps
    linear0 = 25 / 323 * srgb
    linear1 = np.maximum(eps, ((200 * srgb + 11) / (211)))**(12 / 5)
    return np.where(srgb <= 0.04045, linear0, linear1)


import pytest
import numpy as np
import torch
import gsoup
from pathlib import Path

def test_type_conversions():
    test_numpy_bool = np.array([1, 0, 0, 1], dtype=bool)
    test_torch_bool = gsoup.to_torch(test_numpy_bool)

    test_float_from_bool = gsoup.to_float(test_numpy_bool)
    assert test_float_from_bool.dtype == np.float32
    test_8bit_from_bool = gsoup.to_8b(test_numpy_bool)
    assert test_8bit_from_bool.dtype == np.uint8
    test_8bit_from_float = gsoup.to_8b(test_float_from_bool)
    assert test_8bit_from_float.dtype == np.uint8
    test_float_from_8bit = gsoup.to_float(test_8bit_from_float)
    assert test_float_from_8bit.dtype == np.float32
    test_float_from_bool = gsoup.to_float(test_torch_bool)
    assert test_float_from_bool.dtype == torch.float32
    test_8bit_from_bool = gsoup.to_8b(test_torch_bool)
    assert test_8bit_from_bool.dtype == torch.uint8
    test_8bit_from_float = gsoup.to_8b(test_float_from_bool)
    assert test_8bit_from_float.dtype == torch.uint8
    test_float_from_8bit = gsoup.to_float(test_8bit_from_float)
    assert test_float_from_8bit.dtype == torch.float32

def test_broadcast_batch():
    R = np.random.randn(1, 3, 3)
    t = np.random.randn(1, 3)
    R, t = gsoup.broadcast_batch(R, t)
    assert R.shape == (1, 3, 3)
    assert t.shape == (1, 3)
    R = np.random.randn(2, 3, 3)
    t = np.random.randn(2, 3)
    R, t = gsoup.broadcast_batch(R, t)
    assert R.shape == (2, 3, 3)
    assert t.shape == (2, 3)
    R = np.random.randn(2, 3, 3)
    t = np.random.randn(1, 3)
    R, t = gsoup.broadcast_batch(R, t)
    assert R.shape == (2, 3, 3)
    assert t.shape == (2, 3)
    R = np.random.randn(1, 3, 3)
    t = np.random.randn(2, 3)
    R, t = gsoup.broadcast_batch(R, t)
    assert R.shape == (2, 3, 3)
    assert t.shape == (2, 3)
    R = np.random.randn(3)
    t = np.random.randn(3)
    R, t = gsoup.broadcast_batch(R, t)
    assert R.shape == (1, 3)
    assert t.shape == (1, 3)
    R = np.random.randn(1)
    t = np.random.randn(1)
    R, t = gsoup.broadcast_batch(R, t)
    assert R.shape == (1, 1)
    assert t.shape == (1, 1)
    R = np.random.randn(3, 3)
    t = np.random.randn(3)
    with pytest.raises(ValueError):
        R, t = gsoup.broadcast_batch(R, t)

def test_transforms():
    R = np.random.randn(2, 3, 3)
    t = np.random.randn(1, 3)
    Rt = gsoup.compose_rt(R, t)
    assert Rt.shape == (2, 3, 4)

    eye = np.array([1, 0, 0], dtype=np.float32)
    at = np.array([0, 0, 0], dtype=np.float32)
    up = np.array([0, 0, 1], dtype=np.float32)
    transform_np = gsoup.look_at_np(eye, at, up)
    transform_torch = gsoup.look_at_torch(gsoup.to_torch(eye), gsoup.to_torch(at), gsoup.to_torch(up))
    assert np.allclose(transform_np, gsoup.to_numpy(transform_torch))
    transform_np_opengl = gsoup.look_at_np(eye, at, up, opengl=True)
    transform_torch_opengl = gsoup.look_at_torch(gsoup.to_torch(eye), gsoup.to_torch(at), gsoup.to_torch(up), opengl=True)
    assert np.allclose(transform_np_opengl, gsoup.to_numpy(transform_torch_opengl))
    normal = torch.tensor([0, 0, 1.0])

    location3d = np.array([0.1, 0.1, 0.1])
    location3d_noise = location3d + np.random.randn(3) * 0.01
    location4d = gsoup.to_hom(location3d)
    location4d_noise = gsoup.to_hom(location3d_noise)
    # sanity on opengl projection
    v2w = gsoup.look_at_np(np.array([1., 0, 0]), location3d, np.array([0, 0, 1.0]), opengl=True)[0]
    w2v = np.linalg.inv(v2w)
    v2c = gsoup.perspective_projection()
    opengl_location = v2c @ w2v @ location4d
    opengl_location = gsoup.homogenize(opengl_location)
    assert np.allclose(opengl_location[:2], np.zeros(2))
    # sanity on opencv projection
    c2w = gsoup.look_at_np(np.array([1., 0, 0]), location3d, np.array([0, 0, 1.0]))[0]
    w2c = np.linalg.inv(c2w)
    K = gsoup.opencv_intrinsics_from_opengl_project(v2c, 1, 1)
    opencv_location = K @ gsoup.to_34(w2c) @ location4d
    opencv_location = gsoup.homogenize(opencv_location)
    assert np.allclose(opencv_location[:2], np.ones(2)*0.5)
    # test opencv / opengl conversion
    opengl_location = v2c @ w2v @ location4d_noise
    opengl_location = gsoup.homogenize(opengl_location)
    opencv_location = K @ gsoup.to_34(w2c) @ location4d_noise
    opencv_location = gsoup.homogenize(opencv_location)
    # x is in same direction, but opengl screen is -1 to 1 (while opencv is 0 to 1)
    assert np.allclose((opencv_location[0] - opengl_location[0]/2), 0.5)
    # y is in opposite direction, but opengl screen is -1 to 1 (while opencv is 0 to 1)
    assert np.allclose((opencv_location[1] + opengl_location[1]/2), 0.5)

def test_rotations():
    qvecs = gsoup.random_qvec(10)
    torch_qvecs = torch.tensor(qvecs)
    rotmats = gsoup.batch_qvec2mat(qvecs)
    assert rotmats.shape == (10, 3, 3)
    rotmats = gsoup.batch_qvec2mat(torch_qvecs)
    assert rotmats.shape == (10, 3, 3)
    new_qvecs = gsoup.batch_mat2qvec(rotmats)
    mask1 = (torch.abs(new_qvecs - torch_qvecs) < 1e-6)
    mask2 = (torch.abs(new_qvecs + torch_qvecs) < 1e-6)
    assert torch.all(mask1 | mask2)
    rotmat = gsoup.qvec2mat(torch_qvecs[0])
    assert rotmat.shape == (3, 3)
    new_qvec = gsoup.mat2qvec(rotmat)
    mask1 = (torch.abs(new_qvec - torch_qvecs[0]) < 1e-6)
    mask2 = (torch.abs(new_qvec + torch_qvecs[0]) < 1e-6)
    assert torch.all(mask1 | mask2)
    normal = torch.tensor([0, 0, 1.0])
    random_vectors = gsoup.random_vectors_on_sphere(10, normal=normal)
    assert random_vectors.shape == (10, 3)
    assert (random_vectors @ normal).all() > 0
    normal = torch.tensor([[0, 0, 1.0]]).repeat(10, 1)
    random_vectors = gsoup.random_vectors_on_sphere(10, normal=normal)
    assert random_vectors.shape == (10, 3)
    assert (random_vectors[:, None, :] @ normal[:, :, None]).all() > 0
    rotx = gsoup.rotx(np.pi/2, degrees=False)
    assert np.allclose(rotx, np.array([[1., 0, 0, 0], [0, 0, -1, 0], [0, 1, 0, 0], [0, 0, 0, 1]]))

def test_homogenize():
    x = np.random.rand(100, 2)
    hom_x = gsoup.to_hom(x)
    assert hom_x.shape == (100, 3)
    hom_x = gsoup.to_hom(hom_x)
    assert hom_x.shape == (100, 4)
    dehom_x = gsoup.homogenize(hom_x)
    assert dehom_x.shape == (100, 3)
    dehom_x = gsoup.homogenize(dehom_x, keepdim=True)
    assert dehom_x.shape == (100, 3)
    x = np.random.rand(3)
    hom_x = gsoup.to_hom(x)
    assert hom_x.shape == (4,)
    dehom_x = gsoup.homogenize(hom_x)
    assert dehom_x.shape == (3,)

def test_normalize_vertices():
    v = np.random.rand(100, 3) * 100
    v_normalized = gsoup.normalize_vertices(v)
    assert (v_normalized < 1.0).all()

    v = torch.rand(100, 3) * 100
    v_normalized = gsoup.normalize_vertices(v)
    assert (v_normalized < 1.0).all()

def test_structures():
    v, f = gsoup.structures.cube()
    gsoup.save_obj(v, f, "resource/cube.obj")
    v1, f1 = gsoup.load_obj("resource/cube.obj")
    assert np.allclose(v, v1)
    assert np.allclose(f, f1)
    v, f = gsoup.structures.icosehedron()
    gsoup.save_mesh(v, f, "resource/ico.obj")
    v1, f1 = gsoup.load_obj("resource/ico.obj")
    assert np.allclose(v, v1)
    assert np.allclose(f, f1)
    v, f = gsoup.load_obj("resource/cube.obj")
    assert v.shape[0] == 8
    assert f.shape[0] == 12

def test_image():
    checkboard = gsoup.generate_checkerboard(512, 512, 8)
    gsoup.save_image(checkboard, "resource/checkboard.png")
    checkboard_RGBA = np.tile(checkboard, (1, 1, 4))
    checkboard_RGB = gsoup.alpha_compose(checkboard_RGBA, ~checkboard_RGBA[..., :3])
    assert (checkboard_RGB == 1.0).all()
    checkboard_RGB = gsoup.alpha_compose(checkboard_RGBA, bg_color=np.array([0.0, 0.0, 1.0]))
    assert (checkboard_RGB[..., -1] == 1.0).all()
    lollipop_path = Path("resource/lollipop.png")
    lollipop = gsoup.generate_lollipop_pattern(512, 512, dst=lollipop_path)
    gsoup.save_image(lollipop, lollipop_path)
    lollipop2 = gsoup.load_image(lollipop_path)
    assert np.allclose(lollipop, lollipop2)
    gsoup.save_images(lollipop[None, ...], lollipop_path.parent, file_names=["test_save.png"])
    lollipop_pad = gsoup.pad_to_res(lollipop[None, ...], 512, 1024)
    assert lollipop_pad.shape == (1, 512, 1024, 3)
    lollipop_padded_square = gsoup.pad_to_square(lollipop_pad)
    assert lollipop_padded_square.shape == (1, 1024, 1024, 3)
    lollipop_cropped_square = gsoup.crop_to_square(lollipop_pad)
    assert lollipop_cropped_square.shape == (1, 512, 512, 3)
    lollipop_srgb = gsoup.linear_to_srgb(lollipop)
    lollipop_linear = gsoup.srgb_to_linear(lollipop_srgb)
    assert np.allclose(lollipop, lollipop_linear)
    gsoup.generate_concentric_circles(256, 512, dst=Path("resource/circles.png"))
    gsoup.generate_stripe_pattern(256, 512, direction="both", dst=Path("resource/stripe.png"))
    gsoup.generate_dot_pattern(512, 256, dst=Path("resource/dots.png"))
    gray1 = gsoup.generate_gray_gradient(256, 256, grayscale=True, dst=Path("resource/gg_vert.png"))
    assert gray1.shape == (256, 256)
    assert len(np.unique(gray1)) == 10
    assert gray1.max() == 255
    gray2 = gsoup.generate_gray_gradient(50, 800, vertical=False, dst=Path("resource/gg_horiz.png"))
    assert gray2.shape == (50, 800, 3)
    assert gray2.max() == 255
    gray3 = gsoup.generate_gray_gradient(256, 256, bins=-65, dst=Path("resource/gg_bin_min.png"))
    assert gray3.max() == 0
    gray4 = gsoup.generate_gray_gradient(256, 256, bins=300, dst=Path("resource/gg_bin_max.png"))
    assert gray4.max() == 255
    gray5 = gsoup.generate_gray_gradient(1080, 1920, bins=300, dst=Path("resource/gg_highres.png"))
    assert gray5.shape == (1080, 1920, 3)
    assert gray5.max() == 255
    dst = Path("resource/voronoi.png")
    gsoup.generate_voronoi_diagram(512, 512, 1000, dst=dst)
    img = gsoup.load_image(dst)
    assert img.shape == (512, 512, 3)
    assert img.dtype == np.uint8
    img = gsoup.load_image(dst, as_grayscale=True)
    assert img.shape == (512, 512)
    assert img.dtype == np.uint8
    img = gsoup.load_image(dst, to_float=True)
    assert img.shape == (512, 512, 3)
    assert img.dtype == np.float32
    assert (img>=0.0).all()
    assert (img<=1.0).all()
    img = gsoup.load_image(dst, channels_last=False)
    assert img.shape == (3, 512, 512)
    img = gsoup.load_image(dst, to_float=True, as_grayscale=True)
    assert img.shape == (512, 512)
    assert img.dtype == np.float32
    assert (img>=0.0).all()
    assert (img<=1.0).all()
    img = gsoup.load_image(dst, channels_last=False, as_grayscale=True)
    assert img.shape == (512, 512)
    img = gsoup.load_images([dst])
    assert img.shape == (1, 512, 512, 3)
    img = gsoup.load_images([dst], as_grayscale=True)
    assert img.shape == (1, 512, 512)
    img = gsoup.load_images([dst])
    assert img.shape == (1, 512, 512, 3)
    img = gsoup.load_images([dst, dst, dst, dst])
    assert img.shape == (4, 512, 512, 3)
    resized_img = gsoup.resize_images_naive(img, 256, 256, mode="mean")
    assert resized_img.shape == (4, 256, 256, 3)
    resized_img_float = gsoup.resize_images_naive(gsoup.to_float(img), 256, 256, mode="mean")
    assert resized_img_float.shape == (4, 256, 256, 3)
    assert resized_img_float.dtype == np.float32
    resized_img_gray = gsoup.resize_images_naive(img[..., 0:1], 256, 256, mode="mean")
    assert resized_img_gray.shape == (4, 256, 256, 1)
    grid = gsoup.image_grid(resized_img, 2, 2)
    assert grid.shape == (512, 512, 3)
    img = gsoup.load_images([dst, dst, dst, dst], as_grayscale=True)
    assert img.shape == (4, 512, 512)
    img = gsoup.load_images([dst, dst, dst, dst], as_grayscale=True, channels_last=False)
    assert img.shape == (4, 512, 512)
    img = gsoup.load_images([dst, dst, dst, dst], resize_wh=(128, 128))
    assert img.shape == (4, 128, 128, 3)
    img, paths = gsoup.load_images([dst, dst, dst, dst], resize_wh=(128, 256), as_grayscale=True, channels_last=False, return_paths=True, to_float=True, to_torch=True)
    assert len(paths) == 4
    assert img.dtype == torch.float32
    assert img.shape == (4, 256, 128)

def test_video():
    # import os
    # FFMPEG_DIR = os.path.join("/usr/bin")
    # os.environ['PATH'] = FFMPEG_DIR + ":" + os.environ['PATH']
    frame_number = 100
    images = np.random.randint(0, 255, (frame_number, 512, 512, 3), dtype=np.uint8)
    # im1 = gsoup.generate_voronoi_diagram(512, 512, 1000)
    # im2 = gsoup.generate_voronoi_diagram(512, 512, 1000)
    # im1s = np.tile(im1[None, ...], (10, 1, 1, 1))
    # im2s = np.tile(im2[None, ...], (10, 1, 1, 1))
    # images = np.vstack([im1s, im2s])
    dst = Path("resource/noise.avi")
    gsoup.save_video(images, dst, fps=10)
    gsoup.save_video(images, Path("resource/noise_lossy.avi"), lossy=True, fps=10)
    reader = gsoup.VideoReader(dst, h=512, w=512)
    fps = gsoup.FPS()
    for i, frame in enumerate(reader):
        print("{}: {}, fps: {}".format(i, frame.shape, fps()))
        assert np.all(frame == images[i])
    video_frames = gsoup.load_video(dst)
    assert video_frames.shape == (frame_number, 512, 512, 3)
    assert np.all(video_frames == images)
    video_frames_reversed = gsoup.reverse_video(dst)
    assert (video_frames_reversed[-1] == video_frames[0]).all()
    sliced_frames = gsoup.slice_from_video(dst, every_n_frames=2, start_frame=0, end_frame=6)
    assert (sliced_frames == video_frames[:7:2, :, :, :]).all()
    gsoup.video_to_images(dst, Path("resource/noise"))
    gsoup.save_video(Path("resource/noise"), Path("resource/noise2.avi"), fps=10)
    gsoup.save_video(Path("resource/noise"), Path("resource/noise_lossy2.avi"), lossy=True, fps=10)
    discrete_images = gsoup.load_images(Path("resource/noise"))
    assert discrete_images.shape == (frame_number, 512, 512, 3)
    timestamps = gsoup.get_frame_timestamps(dst)
    assert timestamps[0] == 0

def test_procam():
    gray = gsoup.GrayCode()
    patterns = gray.encode((128, 128))
    forward_map, fg = gray.decode(patterns, (128, 128),
                                  output_dir=Path("resource/pix2pix"), mode="ij", debug=True)
    backward_map = gsoup.compute_backward_map((128, 128), forward_map, fg,
                                              output_dir=Path("resource/pix2pix"), debug=True)
    desired = gsoup.generate_lollipop_pattern(128, 128)
    # desired = gsoup.to_float(desired)
    warp_image = gsoup.warp_image(backward_map, desired, cam_wh=(forward_map.shape[1], forward_map.shape[0]),
                                  output_path=Path("resource/warp.png"))
    assert warp_image.shape == (128, 128, 3)
    assert warp_image.dtype == np.uint8
    assert np.mean(np.abs(desired - warp_image)) < 10  # surely an identity corrospondence & warp can't be too bad
    calibration_dir = Path("resource/calibration")
    calibration_dir.mkdir(exist_ok=True, parents=True)
    checkerboard = gsoup.to_float(gsoup.generate_checkerboard(128, 128, 16))
    # T = gsoup.random_perspective()
    # T_opencv = T[:2, :]
    # img_transformed = cv2.warpPerspective(checkerboard, T, (128, 128))
    captures = np.bitwise_and(patterns==255, checkerboard[None, ...]==1.0)
    gsoup.save_images(captures, Path(calibration_dir, "0"))
    gsoup.save_images(captures, Path(calibration_dir, "1"))
    cam_int, cam_dist,\
    proj_int, proj_dist,\
    cam_proj_rmat, cam_proj_tvec = gsoup.calibrate_procam(128, 128, 1, calibration_dir, chess_vert=7, chess_hori=7)

def test_sphere_tracer():
    image_size = 512
    device = "cuda:0"
    w2v, v2c = gsoup.create_random_cameras_on_unit_sphere(5, 1.0, opengl=True, device=device)
    ray_origins, ray_directions = gsoup.generate_rays(w2v, v2c[0], image_size, image_size, device=device)
    sdf = gsoup.structures.sphere_sdf(0.25)
    images = []
    for o, d in zip(ray_origins, ray_directions):
        result = gsoup.render(sdf, o.view(-1, 3), d.view(-1, 3))
        images.append(result.view(image_size, image_size, 4))
    images = gsoup.to_np(torch.stack(images))
    images = gsoup.alpha_compose(images)
    gizmo_images = gsoup.draw_gizmo_on_image(images, gsoup.to_np(v2c @ w2v), opengl=True)
    gsoup.save_images(gizmo_images, Path("resource/sphere_trace"))

def test_qem():
    v, f = gsoup.structures.cube()
    v_new, f_new = gsoup.qem(v, f, budget = 4)
    assert f_new.shape[0] == 4